#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timer.h>
#include <sys/return_code.h>
#include <sys/process.h>
#include <cell/audio.h>
#include <cell/sysmodule.h>
#include <cell/cell_fs.h>
#include <cell/pad.h>
#include <stddef.h>
#include <math.h>
#include <sysutil/sysutil_sysparam.h>

#include "snes9xps3.h"

#include "snes9x.h"
#include "memmap.h"
#include "apu/apu.h"
#include "gfx.h"
#include "snapshot.h"
#include "controls.h"
#include "cheats.h"
#include "movie.h"
#include "logger.h"
#include "display.h"
#include "conffile.h"

#include "cellpadutil.h"
#include "cellgraphicsutil.h"
#include "font.h"

SYS_PROCESS_PARAM(1001, 0x10000);

#define MODE_ROM_MENU 0
#define MODE_SNES9X_START 1
#define MODE_EXIT 2

static bool is_running;
static bool displayinfo = true;
static int mode_switch = MODE_ROM_MENU;
static char rom_path[1024];
static char path[1024];
static bool snes9x_loaded = false;
static bool load_settings = true;
static bool load_rom = false;

int current_save_state;
#define MAX_NO_OF_SETTINGS	6		// if you add more settings to the screen, remember to change this value to the correct number

#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))


static const char *debug_text;

#define NO_JOYSTICKS	0
#define TWO_JOYSTICKS	1
#define MULTITAP	2
#define MOUSE		3
#define SUPERSCOPE	4
static uint8_t controller_settings = NO_JOYSTICKS;


void sysutil_exit_callback (uint64_t status, uint64_t param, void *userdata) {
	(void) param;
	(void) userdata;

	switch (status) {
		case CELL_SYSUTIL_REQUEST_EXITGAME:
			is_running = false;
			mode_switch = MODE_EXIT;
			break;
		case CELL_SYSUTIL_DRAWING_BEGIN:
		case CELL_SYSUTIL_DRAWING_END:
			break;
	}
}





// SNES buttons


#define BTN_A 1
#define BTN_B 2
#define BTN_X 3
#define BTN_Y 4
#define BTN_SELECT 5
#define BTN_START 6
#define BTN_L 7
#define BTN_R 8
#define BTN_LEFT 9
#define BTN_RIGHT 10
#define BTN_UP 11
#define BTN_DOWN 12
#define BTN_QUICKSAVE 13
#define BTN_QUICKLOAD 14

//for mouse
#define MOUSE1_LEFTBUTTON 15
#define MOUSE1_RIGHTBUTTON 16

#define PAD_1 1
#define PAD_2 2
#define PAD_3 3
#define PAD_4 4

#define MAKE_BUTTON(pad, btn) (((pad)<<4)|(btn))

extern color4_t white =		{ 0xFF, 0xFF, 0xFF, 0xFF };
extern color4_t red =		{ 0xFF, 0xFF, 0, 0 };
extern color4_t green =		{ 0xFF, 0, 0xFF, 0 };
extern color4_t black =		{ 0, 0, 0, 0 };
extern color4_t yellow =	{ 0xFF, 0xFF, 0xFF, 0 };

void debugOutChr (int baseX, int baseY, uint8_t chr, color4_t color) {

	int line = ((chr - 32) >> 4) * FONT_HEIGHT;
	int offset = ((chr - 32) & 15) * FONT_WIDTH;
	color4_t *ptr = &display[baseY * color_buf[0].width + baseX];

	for (int h = 0; h < FONT_HEIGHT * FONT_SCALE; ++h, ptr += color_buf[0].width - FONT_WIDTH * FONT_SCALE) {
		for (int w = 0; w < FONT_WIDTH * FONT_SCALE; ++w, ++ptr) {
			char p = font[line + h / FONT_SCALE][offset + w / FONT_SCALE];

			if (p == '#')
				*ptr = color;
			else if (p == '.')
				*ptr = black;
		}
	}
}

void debugOutStr (int baseX, int baseY, const char *str, color4_t color) {
	for (const char *ptr = str; *ptr; ++ptr) {
		debugOutChr(baseX, baseY, *ptr, color);
		baseX += FONT_WIDTH * FONT_SCALE;
	}
}

void debugOutStr (int baseX, int baseY, const char *str) {
	debugOutStr(baseX, baseY, str, white);
}

bool init_settings () {
	ZeroMemory(&Settings, sizeof(Settings));

	Settings.MouseMaster = TRUE;
	Settings.SuperScopeMaster = TRUE;
	Settings.JustifierMaster = TRUE;
	Settings.MultiPlayer5Master = TRUE;
	Settings.FrameTimePAL = 20000;
	Settings.FrameTimeNTSC = 16667;
	Settings.SixteenBitSound = TRUE;
	Settings.Stereo = TRUE;
	Settings.SoundPlaybackRate = 48000;
	Settings.SoundInputRate = 31942;
	Settings.SoundSync = TRUE;
	Settings.SupportHiRes = TRUE;
	Settings.Transparency = TRUE;
	Settings.AutoDisplayMessages = TRUE;
	Settings.InitialInfoStringTimeout = 120;
	Settings.HDMATimingHack = 100;
	Settings.BlockInvalidVRAMAccessMaster = TRUE;
	Settings.StopEmulation = TRUE;
	Settings.WrongMovieStateProtection = TRUE;
	Settings.DumpStreamsMaxFrames = -1;
	Settings.StretchScreenshots = 1;
	Settings.SnapshotScreenshots = TRUE;
	Settings.SkipFrames = AUTO_FRAMERATE;
	Settings.TurboSkipFrames = 15;
	Settings.CartAName[0] = 0;
	Settings.CartBName[0] = 0;
	Settings.AutoSaveDelay = 0;
	Settings.DisplayFrameRate = FALSE;
	Settings.DisableGraphicWindows = FALSE;

	CPU.Flags = 0;

}

bool init_snes9x()
{
	if(snes9x_loaded)
	{
		Memory.Deinit();
	}
	Memory.Init();

	if(snes9x_loaded)
	{
		S9xDeinitAPU();
	}
	S9xInitAPU();

	S9xInitSound(100, 25);

	S9xSetSoundMute(TRUE);
	S9xSetRenderPixelFormat(RGB565);

	if (!Memory.LoadROM(rom_path)) {
		debug_text = "load rom failed";
		return false;
	}

	Memory.LoadSRAM(S9xGetFilename(".srm", SRAM_DIR));

	GFX.Pitch = 1024;
	GFX.Screen = (uint16_t *)memalign(1024, 1024 * 478);

	if(snes9x_loaded)
	{
		S9xGraphicsDeinit();
	}
	S9xGraphicsInit();

	// Controller initialization
	switch(controller_settings)
	{
		case TWO_JOYSTICKS:
			S9xSetController(0, CTL_JOYPAD, 0, 0, 0, 0);
			S9xSetController(1, CTL_JOYPAD, 1, 0, 0, 0);
			break;
		case MULTITAP:
			S9xSetController(0, CTL_JOYPAD,	0,0,0,0);
			S9xSetController(1, CTL_MP5,	1,2,3,4);
			break;
	}

	S9xUnmapAllControls();

#define MAP_BUTTON(id, name) S9xMapButton((id), S9xGetCommandT((name)), false)

	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_A), "Joypad1 A");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_B), "Joypad1 B");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_X), "Joypad1 X");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_Y), "Joypad1 Y");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_SELECT), "Joypad1 Select");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_START), "Joypad1 Start");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_L), "Joypad1 L");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_R), "Joypad1 R");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_LEFT), "Joypad1 Left");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_RIGHT), "Joypad1 Right");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_UP), "Joypad1 Up");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_DOWN), "Joypad1 Down");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_QUICKSAVE), "SaveFreezeFile");
	MAP_BUTTON(MAKE_BUTTON(PAD_1, BTN_QUICKLOAD), "LoadFreezeFile");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_A), "Joypad2 A");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_B), "Joypad2 B");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_X), "Joypad2 X");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_Y), "Joypad2 Y");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_SELECT), "Joypad2 Select");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_START), "Joypad2 Start");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_L), "Joypad2 L");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_R), "Joypad2 R");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_LEFT), "Joypad2 Left");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_RIGHT), "Joypad2 Right");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_UP), "Joypad2 Up");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_DOWN), "Joypad2 Down");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_QUICKSAVE), "SaveFreezeFile");
	MAP_BUTTON(MAKE_BUTTON(PAD_2, BTN_QUICKLOAD), "LoadFreezeFile");
	switch(controller_settings)
	{
		case MULTITAP:
			//player 3 - for multitap
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_A), "Joypad3 A");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_B), "Joypad3 B");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_X), "Joypad3 X");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_Y), "Joypad3 Y");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_SELECT), "Joypad3 Select");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_START), "Joypad3 Start");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_L), "Joypad3 L");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_R), "Joypad3 R");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_LEFT), "Joypad3 Left");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_RIGHT), "Joypad3 Right");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_UP), "Joypad3 Up");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_DOWN), "Joypad3 Down");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_QUICKSAVE), "SaveFreezeFile");
			MAP_BUTTON(MAKE_BUTTON(PAD_3, BTN_QUICKLOAD), "LoadFreezeFile");
			//player 4 - for multitap
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_A), "Joypad4 A");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_B), "Joypad4 B");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_X), "Joypad4 X");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_Y), "Joypad4 Y");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_SELECT), "Joypad4 Select");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_START), "Joypad4 Start");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_L), "Joypad4 L");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_R), "Joypad4 R");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_LEFT), "Joypad4 Left");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_RIGHT), "Joypad4 Right");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_UP), "Joypad4 Up");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_DOWN), "Joypad4 Down");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_QUICKSAVE), "SaveFreezeFile");
			MAP_BUTTON(MAKE_BUTTON(PAD_4, BTN_QUICKLOAD), "LoadFreezeFile");
			break;
	}
	
	return true;
}


void read_pad () {
	uint8_t pads_connected = cellPadUtilNumberPadsConnected();
	for (uint8_t i = 0; i < pads_connected; ++i) {
		if (cellPadUtilUpdate(i))
		{
			uint8_t pad_id = i + 1;
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_A), (cellPadUtilButtonPressed(i,CTRL_CIRCLE)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_B), (cellPadUtilButtonPressed(i,CTRL_CROSS)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_X), (cellPadUtilButtonPressed(i,CTRL_TRIANGLE)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_Y), (cellPadUtilButtonPressed(i,CTRL_SQUARE)));

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_SELECT), (cellPadUtilButtonPressed(i,CTRL_SELECT)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_START), (cellPadUtilButtonPressed(i,CTRL_START)));

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_L), (cellPadUtilButtonPressed(i,CTRL_L1)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_R), (cellPadUtilButtonPressed(i,CTRL_R1)));

			S9xReportButton(MAKE_BUTTON(pad_id, BTN_UP), (cellPadUtilButtonPressed(i,CTRL_UP) | cellPadUtilAnalogPressedUp(i,CTRL_LSTICK)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_DOWN), (cellPadUtilButtonPressed(i,CTRL_DOWN) | cellPadUtilAnalogPressedDown(i,CTRL_LSTICK)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_LEFT), (cellPadUtilButtonPressed(i,CTRL_LEFT) | cellPadUtilAnalogPressedLeft(i,CTRL_LSTICK)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_RIGHT), (cellPadUtilButtonPressed(i,CTRL_RIGHT) | cellPadUtilAnalogPressedRight(i,CTRL_LSTICK)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_QUICKSAVE), (cellPadUtilButtonPressed(i,CTRL_R2) & cellPadUtilButtonPressed(i,CTRL_R3)));
			S9xReportButton(MAKE_BUTTON(pad_id, BTN_QUICKLOAD), (cellPadUtilButtonPressed(i,CTRL_L2) & cellPadUtilButtonPressed(i,CTRL_R3)));
			if(cellPadUtilButtonPressed(i,CTRL_R3) && cellPadUtilButtonPressed(i,CTRL_L3))
			{
				is_running = false;
				mode_switch = MODE_ROM_MENU;
			}
			
		}
	}
}

#define AUDIO_BLOCK_COUNT 16 
#define AUDIO_CHANNEL_COUNT 2

static uint32_t audio_port;
static float *audio_buffer;
static float *audio_buffer_end;
sys_addr_t audio_read_index_addr;
static int audio_block_index;
static int audio_last_read_block;

void init_sound () {
	CellAudioPortParam params;
	CellAudioPortConfig config;

	cellAudioInit();

	params.nChannel = AUDIO_CHANNEL_COUNT;
	params.nBlock = AUDIO_BLOCK_COUNT;
	params.attr = 0;

	if (cellAudioPortOpen(&params, &audio_port) != CELL_OK) {
		cellAudioQuit();
		return;
	}

	cellAudioGetPortConfig(audio_port, &config);
	audio_buffer = (float *)config.portAddr;
	audio_buffer_end = (float *)(config.portAddr + config.portSize);
	audio_read_index_addr = config.readIndexAddr;

	cellAudioPortStart(audio_port);
}

void make_noise () {
	int16_t samples[256 * 2 * 2];

	if (S9xGetSampleCount() < 512)
		return;

	int current_block = *(uint64_t *)audio_read_index_addr;

	if (current_block == audio_last_read_block)
		return;

	while (S9xGetSampleCount() >= 512 && audio_block_index != current_block) {
		S9xMixSamples((uint8 *)samples, 256 * 2);

		float *buf = audio_buffer + AUDIO_CHANNEL_COUNT * 256 * audio_block_index;
		for (int i = 0; i < 256 * AUDIO_CHANNEL_COUNT; ++i)
			buf[i] = ((float)samples[i])/32768.0f;

		audio_block_index = (audio_block_index + 1) % AUDIO_BLOCK_COUNT;
	}

	audio_last_read_block = current_block;
}

// void do_settings()
// // called from ROM menu by pressing the SELECT button
// // return to ROM menu by pressing the CIRCLE button
void do_settings()
{
	int currently_selected_setting = 0;
	char sound_input_buf[10], frameskip_buf[10], save_state_char[10];

	while (is_running) {

		if(cellPadUtilUpdate(0)) {
				if (cellPadUtilButtonPressed(0, CTRL_CIRCLE)) // back to ROM menu if CIRCLE is pressed
				return;

				if (cellPadUtilButtonPressed(0, CTRL_DOWN) | cellPadUtilAnalogPressedDown(0, CTRL_LSTICK))	// down to next setting
					if(currently_selected_setting < MAX_NO_OF_SETTINGS-1)	// if you add more settings, remember to change MAX_NO_OF_SETTINGS
						currently_selected_setting++;

				if (cellPadUtilButtonPressed(0, CTRL_UP) | cellPadUtilAnalogPressedUp(0, CTRL_LSTICK))	// up to previous setting
					if(currently_selected_setting > 0)
						currently_selected_setting--;

				if (cellPadUtilButtonPressed(0,CTRL_LEFT) | cellPadUtilButtonPressed(0, CTRL_RIGHT) | cellPadUtilAnalogPressedLeft(0,CTRL_LSTICK) | cellPadUtilAnalogPressedRight(0,CTRL_LSTICK)) {
						switch(currently_selected_setting) {

						// display framerate on/off
						case 0:
							Settings.DisplayFrameRate = !Settings.DisplayFrameRate;
							break;

						// sound input rate
						// Playing with this setting in the emulator doesn't seem to have any effect on the sound output.
						// Maybe we need to re-initialise the audio subsystem here for the changes to have any effect.
						// I don't know whether it is possible to re-initialise the audio engine on its own or whether we
						// need to re-initialize the entire emulator.
						case 1:
							if(cellPadUtilButtonPressed(0, CTRL_LEFT) | cellPadUtilAnalogPressedLeft(0,CTRL_LSTICK))
								if(Settings.SoundInputRate > 30000)
									Settings.SoundInputRate--;
							if(cellPadUtilButtonPressed(0, CTRL_RIGHT) | cellPadUtilAnalogPressedRight(0,CTRL_LSTICK))
								if(Settings.SoundInputRate < 32000) // SNES APU has an input rate of 32000, no higher
									Settings.SoundInputRate++;

							break;

						// transparency on/off - works fine
						case 2:
							Settings.Transparency = !Settings.Transparency;
							break;

						//frameskip
						//Changing this value seems to ahve no effect in the emulator
						//I can't find any reference to the SkipFrames value in the emulator engine itself
						//so even though it is a setting, the emulator engine seems to blatantly ignore it
						//This is confirmed by compiling with a value of skipFrames = 5, it makes no difference to to the output
						case 3:
						       if(cellPadUtilButtonPressed(0, CTRL_LEFT) | cellPadUtilAnalogPressedLeft(0,CTRL_LSTICK))
						       if (Settings.SkipFrames <= 1)
								Settings.SkipFrames = AUTO_FRAMERATE;
						       else if (Settings.SkipFrames != AUTO_FRAMERATE)
								Settings.SkipFrames--;

						       if(cellPadUtilButtonPressed(0, CTRL_RIGHT) | cellPadUtilAnalogPressedRight(0,CTRL_LSTICK))
							       if (Settings.SkipFrames == AUTO_FRAMERATE)
								       Settings.SkipFrames = 1;
						       	else if (Settings.SkipFrames < 10)
								Settings.SkipFrames++;
						       break;

						//disable graphic windows on/off
						//doesn't seem to do anything but maybe I just haven't played any games which have
						//graphic windows
						case 4:
						       Settings.DisableGraphicWindows = !Settings.DisableGraphicWindows;
						       break;
						// change current save state slot
						case 5:
						       if(cellPadUtilButtonPressed(0, CTRL_LEFT) | cellPadUtilAnalogPressedLeft(0,CTRL_LSTICK))
							       if(current_save_state > 0)
								       current_save_state--;
						       if(cellPadUtilButtonPressed(0, CTRL_RIGHT) | cellPadUtilAnalogPressedRight(0,CTRL_LSTICK))
							       if(current_save_state < 9)
								       current_save_state++;
						       break;
						} // end of switch

				}
		}
		// clear screen to black
		memset(display, 0, color_buf[0].pitch * color_buf[0].height);

		debugOutStr(32, 32, "Display framerate");
		debugOutStr(280, 32, Settings.DisplayFrameRate == TRUE ? "ON" : "OFF", currently_selected_setting == 0 ? yellow : red);

		debugOutStr(32, 42, "Sound Input Rate");
		sprintf(sound_input_buf, "%d", Settings.SoundInputRate);
		debugOutStr(280, 42, sound_input_buf, currently_selected_setting == 1 ? yellow : red);

		debugOutStr(32, 52, "Transparency effects");
		debugOutStr(280, 52, Settings.Transparency == TRUE ? "ON" : "OFF", currently_selected_setting == 2 ? yellow : red);

		debugOutStr(32, 62, "Frame skip");
		if(Settings.SkipFrames == AUTO_FRAMERATE)
			debugOutStr(280, 62, "AUTO", currently_selected_setting == 3 ? yellow : red);
		else {
			sprintf(frameskip_buf, "%d", Settings.SkipFrames);
			debugOutStr(280, 62, frameskip_buf, currently_selected_setting == 3 ? yellow : red);
		}

		debugOutStr(32, 72, "Disable graphic windows");
		debugOutStr(280, 72, Settings.DisableGraphicWindows == TRUE ? "ON" : "OFF", currently_selected_setting == 4 ? yellow : red);

		debugOutStr(32, 82, "Current save state slot");
		sprintf(save_state_char,"%d",current_save_state);
		debugOutStr(280, 82, save_state_char, currently_selected_setting == 5 ? yellow : red);

		debugOutStr(370, 430, "UP   / DOWN  - select setting");
		debugOutStr(370, 440, "LEFT / RIGHT - change setting");
		debugOutStr(370, 450, "CIRCLE       - return to ROM menu");

		cellGcmUtilRender(true);
	}
}

void qsort_structs(CellFsDirent entries[], int left, int right)
{
	char *tmp_name;
	CellFsDirent temp_struct;
	register int x, y;

	x = left;
	y = right;
	tmp_name = entries[(left+right)/2].d_name;

	do {
		while((strcasecmp(entries[x].d_name, tmp_name) < 0) && (x < right))
			x++;
		while((strcasecmp(entries[y].d_name, tmp_name) > 0) && (y > left))
			y--;

		if(x <= y) {
			temp_struct = entries[x];
			entries[x] = entries[y];
			entries[y] = temp_struct;
			x++;
			y--;
		}
	} while (x <= y);

	if(left < y)
		qsort_structs(entries, left, y);

	if(x < right)
		qsort_structs(entries, x, right);
}


void rom_menu () {
	bool needs_update = true;
	static CellFsDirent dirent[8192];
	const char *msg = NULL;
	int fd;
	int num_per_page = 47, current_index = 0, file_count = 0;

	//don't go back all the way to the root after quitting a ROM
	if(path[0] == '\0')
	{
		strcpy(path, "/");
	}
	
	while (is_running) {
		if (needs_update) {
			needs_update = false;
			current_index = 0;

			int i = 0;
			if (cellFsOpendir(path, &fd) == CELL_FS_SUCCEEDED) {
				uint64_t nread;

				while (cellFsReaddir(fd, &dirent[i], &nread) == CELL_FS_SUCCEEDED) {
					if (nread == 0)
						break;

					if (dirent[i].d_type != CELL_FS_TYPE_DIRECTORY &&
					    dirent[i].d_type != CELL_FS_TYPE_REGULAR)
						continue; // so lazy.

					if (dirent[i].d_type == CELL_FS_TYPE_REGULAR) {
						int namelen = strlen(dirent[i].d_name);
						if (namelen < 4)
							continue;

						char *ext = dirent[i].d_name + namelen - 4;

						if (strcasecmp(ext, ".smc") && strcasecmp(ext, ".fig") && strcasecmp(ext, ".sfc"))
							continue;
					}

					++i;

					if (i == 8192)
						break;
				}

				cellFsClosedir(fd);
			} else {
				msg = "Directory not found. Try again!";
			}

			file_count = i;

			if (file_count == 0)
				msg = "No ROMs. =(";

			needs_update = false;
		}

		// sort directory file entries - courtesy of cmonkey69
		// start sorting from first entry after '..' and sort file_count - 3 items (indexing starts at zero, don't sort '.' and '..')
		qsort_structs(&dirent[2], 0, file_count - 3);

		if(cellPadUtilUpdate(0)) {
			if (cellPadUtilButtonPressed(0,CTRL_DOWN) | cellPadUtilAnalogPressedDown(0,CTRL_LSTICK))
			{
				current_index = MIN(current_index + 1, file_count - 1);
			}
			if (cellPadUtilButtonPressed(0,CTRL_UP) | cellPadUtilAnalogPressedUp(0,CTRL_LSTICK))
			{
				current_index = MAX(current_index - 1, 0);
			}
			if (cellPadUtilButtonPressed(0,CTRL_RIGHT) | cellPadUtilAnalogPressedRight(0,CTRL_LSTICK))
			{
				current_index = MIN(current_index + 5, file_count - 1);
			}
			if (cellPadUtilButtonPressed(0,CTRL_LEFT) | cellPadUtilAnalogPressedLeft(0,CTRL_LSTICK))
			{
				current_index = MAX(current_index - 5, 0);
			}
			if (cellPadUtilButtonPressed(0,CTRL_R1))
			{
				current_index = MIN(current_index + num_per_page, file_count - 1);
			}
			if (cellPadUtilButtonPressed(0,CTRL_L1))
			{
				current_index = MAX(current_index - num_per_page, 0);
			}
			if (cellPadUtilButtonPressed(0,CTRL_SELECT))
			{
				do_settings();	// jump to settings screen
				needs_update = true;
			}
			if (cellPadUtilButtonPressed(0,CTRL_CROSS))
			{
				displayinfo = false;
				if(dirent[current_index].d_type == CELL_FS_TYPE_DIRECTORY) {
					sprintf(path, "%s/%s", path, dirent[current_index].d_name);
					needs_update = true;
					continue;
				} else {
					//load game (standard controls), go back to main loop
					sprintf(rom_path, "%s/%s", path, dirent[current_index].d_name);
					is_running = false;
					mode_switch = MODE_SNES9X_START;
					if(controller_settings == TWO_JOYSTICKS)
					{
						load_rom = true;
					}
					else
					{
						controller_settings = TWO_JOYSTICKS;
						load_rom = true;
					}
					return;
				}
			}
			if (cellPadUtilButtonPressed(0,CTRL_TRIANGLE))
			{
				displayinfo = false;
				if (dirent[current_index].d_type == CELL_FS_TYPE_DIRECTORY) {
					sprintf(path, "%s/%s", path, dirent[current_index].d_name);
					needs_update = true;
					continue;
				} else {
					//load game with multitap support
					sprintf(rom_path, "%s/%s", path, dirent[current_index].d_name);
					is_running = false;
					mode_switch = MODE_SNES9X_START;
					if(controller_settings == MULTITAP)
					{
						load_rom = true;
					}
					else
					{
						controller_settings = MULTITAP;
						load_rom = true;
					}
					return;
				}
			}
			if (cellPadUtilButtonPressed(0,CTRL_L2) && cellPadUtilButtonPressed(0,CTRL_R2))
			{
				if (snes9x_loaded)
				{
					load_rom = false;
					mode_switch = MODE_SNES9X_START;
				}
				return;
			}
		}
		


		memset(display, 0, color_buf[0].pitch * color_buf[0].height);

		if(displayinfo)
		{
			debugOutStr(400, 420, "CROSS    - Enter Dir/Load ROM");
			debugOutStr(400, 430, "TRIANGLE - Enter Dir/Load ROM w/ multitap");
			debugOutStr(400, 440, "L2 + R2  - Return to game");
			debugOutStr(400, 450, "SELECT   - Settings screen");
		}

		int page_number = current_index / num_per_page;
		int page_base = page_number * num_per_page;

		for (int i = page_base; i < file_count && i < page_base + num_per_page; ++i) {

			// comment out line below if you don't want a 'D' or an 'F' in front of each entry to indicate whether it's a directory or a file
			//debugOutChr(16, 32 + FONT_HEIGHT * FONT_SCALE * (i - page_base), dirent[i].d_type == CELL_FS_TYPE_DIRECTORY ? 'D' : 'F', white);
			
			debugOutStr(32, 32 + FONT_HEIGHT * FONT_SCALE * (i - page_base), dirent[i].d_name, 
					i == current_index ? red : dirent[i].d_type == CELL_FS_TYPE_DIRECTORY ? green : white);
		}

		if (msg)
			debugOutStr(32, 32 + FONT_HEIGHT * FONT_SCALE * (num_per_page + 1), msg);

		cellGcmUtilRender(true);
	}
}

void start_snes9x()
{
	if(load_settings)
	{
		init_settings();
		load_settings = false;
	}

	if(load_rom)
	{
		if (!(snes9x_loaded = init_snes9x())) Settings.Paused = TRUE;
	}

	
	memset(display, 0, color_buf[0].height * color_buf[0].pitch);
	memset(display_back, 0, color_buf[0].height * color_buf[0].pitch);
	S9xSetSoundMute(FALSE);


	while (is_running) {
		read_pad();

		if (!Settings.Paused) {
			make_noise();
			if (S9xSyncSound())
				S9xMainLoop();
		}

		/*
		if (debug_text)
			debugOutStr(100, 100, debug_text);
			*/

		cellGcmUtilRender(false);
	}


	if (snes9x_loaded)
		Memory.SaveSRAM(S9xGetFilename(".srm", SRAM_DIR));
	S9xSetSoundMute(TRUE);
	is_running = true;

}

int main (void) {
	cellSysmoduleLoadModule(CELL_SYSMODULE_FS);
	cellSysmoduleLoadModule(CELL_SYSMODULE_IO);
	cellSysmoduleLoadModule(CELL_SYSMODULE_AUDIO);

	cellSysutilRegisterCallback(0, sysutil_exit_callback, NULL);

	cellPadUtilPadInit();
	cellGcmUtilInitDisplay();
	cellGcmUtilInitShader();
	is_running = true;
	cellGcmUtilInitDisplay2();
	init_sound();

	// load SNES9x settings before calling rom_menu() so they are displayed
	// correctly, in case settings screen is accessed before a ROM is loaded
	current_save_state = 0;
	if (load_settings)
	{
		init_settings();
		load_settings =  false;
	}

	while(1)
	{
		switch(mode_switch)
		{
			case MODE_ROM_MENU:
				rom_menu();
				break;
			case MODE_SNES9X_START:
				//start SNES9x in NTSC mode
				start_snes9x();
				break;
			case MODE_EXIT:
				cellSysmoduleUnloadModule(CELL_SYSMODULE_FS);
				cellSysmoduleUnloadModule(CELL_SYSMODULE_IO);
				cellSysmoduleUnloadModule(CELL_SYSMODULE_AUDIO);
				cellSysutilUnregisterCallback(0);
				return 0;
		}
	}

	return 0;
}

